@@ -14,7 +14,6 @@
 from collections import OrderedDict
 
 import flask
-from six import iteritems, string_types
 
 
 DENY = 'DENY'
@@ -26,8 +25,11 @@ DEFAULT_REFERRER_POLICY = 'strict-origin-when-cross-origin'
 
 DEFAULT_CSP_POLICY = {
     'default-src': '\'self\'',
+    'object-src': '\'none\'',
 }
 
+DEFAULT_SESSION_COOKIE_SAMESITE = "Lax"
+
 GOOGLE_CSP_POLICY = {
     # Fonts from fonts.google.com
     'font-src': '\'self\' themes.googleusercontent.com *.gstatic.com',
@@ -39,13 +41,20 @@ GOOGLE_CSP_POLICY = {
     # Used by generated code from http://www.google.com/fonts
     'style-src': '\'self\' ajax.googleapis.com fonts.googleapis.com '
                  '*.gstatic.com',
+    'object-src': '\'none\'',
     'default-src': '\'self\' *.gstatic.com',
 }
 
+DEFAULT_PERMISSIONS_POLICY = {
+}
+
+DEFAULT_DOCUMENT_POLICY = {
+}
+
 DEFAULT_FEATURE_POLICY = {
 }
 
-NONCE_LENGTH = 16
+NONCE_LENGTH = 32
 
 
 class Talisman(object):
@@ -61,6 +70,8 @@ class Talisman(object):
             self,
             app,
             feature_policy=DEFAULT_FEATURE_POLICY,
+            permissions_policy=DEFAULT_PERMISSIONS_POLICY,
+            document_policy=DEFAULT_DOCUMENT_POLICY,
             force_https=True,
             force_https_permanent=False,
             force_file_save=False,
@@ -74,10 +85,12 @@ class Talisman(object):
             content_security_policy_report_uri=None,
             content_security_policy_report_only=False,
             content_security_policy_nonce_in=None,
-            legacy_content_security_policy_header=True,
             referrer_policy=DEFAULT_REFERRER_POLICY,
             session_cookie_secure=True,
-            session_cookie_http_only=True):
+            session_cookie_http_only=True,
+            session_cookie_samesite=DEFAULT_SESSION_COOKIE_SAMESITE,
+            x_content_type_options=True,
+            x_xss_protection=True):
         """
         Initialization.
 
@@ -85,6 +98,10 @@ class Talisman(object):
             app: A Flask application.
             feature_policy: A string or dictionary describing the
                 feature policy for the response.
+            permissions_policy: A string or dictionary describing the
+                permissions policy for the response.
+            document_policy: A string or dictionary describing the
+                document policy for the response.
             force_https: Redirects non-http requests to https, disabled in
                 debug mode.
             force_https_permanent: Uses 301 instead of 302 redirects.
@@ -97,7 +114,7 @@ class Talisman(object):
                 https://hstspreload.org.
             strict_transport_security_max_age: How long HSTS headers are
                 honored by the browser.
-            strict_transport_security_include_subdomain: Whether to include
+            strict_transport_security_include_subdomains: Whether to include
                 all subdomains when setting HSTS.
             content_security_policy: A string or dictionary describing the
                 content security policy for the response.
@@ -109,15 +126,18 @@ class Talisman(object):
                 the POST data
             content_security_policy_nonce_in: A list of csp sections to include
                 a per-request nonce value in
-            legacy_content_security_policy_header: Whether to add X-CSP header
             referrer_policy: A string describing the referrer policy for the
                 response.
             session_cookie_secure: Forces the session cookie to only be sent
                 over https. Disabled in debug mode.
             session_cookie_http_only: Prevents JavaScript from reading the
                 session cookie.
+            session_cookie_samesite: Sets samesite parameter on session cookie
             force_file_save: Prevents the user from opening a file download
                 directly on >= IE 8
+            x_content_type_options: Prevents MIME type sniffing
+            x_xss_protection: Prevents the page from loading when the browser
+                detects reflected cross-site scripting attacks
 
         See README.rst for a detailed description of each option.
         """
@@ -125,6 +145,17 @@ class Talisman(object):
             self.feature_policy = OrderedDict(feature_policy)
         else:
             self.feature_policy = feature_policy
+
+        if isinstance(permissions_policy, dict):
+            self.permissions_policy = OrderedDict(permissions_policy)
+        else:
+            self.permissions_policy = permissions_policy
+
+        if isinstance(document_policy, dict):
+            self.document_policy = OrderedDict(document_policy)
+        else:
+            self.document_policy = document_policy
+
         self.force_https = force_https
         self.force_https_permanent = force_https_permanent
 
@@ -159,18 +190,21 @@ class Talisman(object):
 
         app.jinja_env.globals['csp_nonce'] = self._get_nonce
 
-        self.legacy_content_security_policy_header = \
-            legacy_content_security_policy_header
-
         self.referrer_policy = referrer_policy
 
         self.session_cookie_secure = session_cookie_secure
 
+        app.config['SESSION_COOKIE_SAMESITE'] = session_cookie_samesite
+
         if session_cookie_http_only:
             app.config['SESSION_COOKIE_HTTPONLY'] = True
 
         self.force_file_save = force_file_save
 
+        self.x_content_type_options = x_content_type_options
+
+        self.x_xss_protection = x_xss_protection
+
         self.app = app
 
         app.before_request(self._force_https)
@@ -189,6 +223,13 @@ class Talisman(object):
             'frame_options_allow_from', self.frame_options_allow_from)
         view_options.setdefault(
             'content_security_policy', self.content_security_policy)
+        view_options.setdefault(
+            'content_security_policy_nonce_in',
+            self.content_security_policy_nonce_in)
+        view_options.setdefault(
+            'permissions_policy', self.permissions_policy)
+        view_options.setdefault(
+            'document_policy', self.document_policy)
         view_options.setdefault(
             'feature_policy', self.feature_policy
         )
@@ -225,7 +266,9 @@ class Talisman(object):
     def _set_response_headers(self, response):
         """Applies all configured headers to the given response."""
         options = self._get_local_options()
-        self._set_feature_headers(response.headers, options)
+        self._set_feature_policy_headers(response.headers, options)
+        self._set_permissions_policy_headers(response.headers, options)
+        self._set_document_policy_headers(response.headers, options)
         self._set_frame_options_headers(response.headers, options)
         self._set_content_security_policy_headers(response.headers, options)
         self._set_hsts_headers(response.headers)
@@ -236,15 +279,30 @@ class Talisman(object):
         local_options = self._get_local_options()
         if (
                 local_options['content_security_policy'] and
-                self.content_security_policy_nonce_in and
+                local_options['content_security_policy_nonce_in'] and
                 not getattr(flask.request, 'csp_nonce', None)):
             flask.request.csp_nonce = get_random_string(NONCE_LENGTH)
 
     def _get_nonce(self):
         return getattr(flask.request, 'csp_nonce', '')
 
+    def _parse_structured_header_policy(self, policy):
+        if isinstance(policy, str):
+            return policy
+
+        policies = []
+        for section, content in policy.items():
+            policy_part = '{}={}'.format(section, content)
+
+            policies.append(policy_part)
+
+        policy = ', '.join(policies)
+
+        return policy
+
     def _parse_policy(self, policy):
-        if isinstance(policy, string_types):
+        local_options = self._get_local_options()
+        if isinstance(policy, str):
             # parse the string into a policy dict
             policy_string = policy
             policy = OrderedDict()
@@ -254,14 +312,14 @@ class Talisman(object):
                 policy[policy_parts[0]] = " ".join(policy_parts[1:])
 
         policies = []
-        for section, content in iteritems(policy):
-            if not isinstance(content, string_types):
+        for section, content in policy.items():
+            if not isinstance(content, str):
                 content = ' '.join(content)
             policy_part = '{} {}'.format(section, content)
 
             if (
                     hasattr(flask.request, 'csp_nonce') and
-                    section in self.content_security_policy_nonce_in):
+                    section in local_options['content_security_policy_nonce_in']):
                 policy_part += " 'nonce-{}'".format(flask.request.csp_nonce)
 
             policies.append(policy_part)
@@ -270,7 +328,7 @@ class Talisman(object):
 
         return policy
 
-    def _set_feature_headers(self, headers, options):
+    def _set_feature_policy_headers(self, headers, options):
         if not options['feature_policy']:
             return
 
@@ -279,6 +337,24 @@ class Talisman(object):
 
         headers['Feature-Policy'] = policy
 
+    def _set_permissions_policy_headers(self, headers, options):
+        if not options['permissions_policy']:
+            return
+
+        policy = options['permissions_policy']
+        policy = self._parse_structured_header_policy(policy)
+
+        headers['Permissions-Policy'] = policy
+
+    def _set_document_policy_headers(self, headers, options):
+        if not options['document_policy']:
+            return
+
+        policy = options['document_policy']
+        policy = self._parse_structured_header_policy(policy)
+
+        headers['Document-Policy'] = policy
+
     def _set_frame_options_headers(self, headers, options):
         if not options['frame_options']:
             return
@@ -289,8 +365,11 @@ class Talisman(object):
                 options['frame_options_allow_from'])
 
     def _set_content_security_policy_headers(self, headers, options):
-        headers['X-XSS-Protection'] = '1; mode=block'
-        headers['X-Content-Type-Options'] = 'nosniff'
+        if self.x_xss_protection:
+            headers['X-XSS-Protection'] = '1; mode=block'
+
+        if self.x_content_type_options:
+            headers['X-Content-Type-Options'] = 'nosniff'
 
         if self.force_file_save:
             headers['X-Download-Options'] = 'noopen'
@@ -310,9 +389,6 @@ class Talisman(object):
             csp_header += '-Report-Only'
 
         headers[csp_header] = policy
-        # IE 10-11, Older Firefox.
-        if self.legacy_content_security_policy_header:
-            headers['X-' + csp_header] = policy
 
     def _set_hsts_headers(self, headers):
         criteria = [
@@ -339,8 +415,9 @@ class Talisman(object):
         """Use talisman as a decorator to configure options for a particular
         view.
 
-        Only frame_options, frame_options_allow_from, and
-        content_security_policy can be set on a per-view basis.
+        Only force_https, frame_options, frame_options_allow_from,
+        content_security_policy, content_security_policy_nonce_in
+        and feature_policy can be set on a per-view basis.
 
         Example:
 
@@ -364,7 +441,12 @@ class Talisman(object):
 
 try:
     import secrets
-    get_random_string = secrets.token_urlsafe  # pragma: no cover
+
+    def get_random_string(length):  # pragma: no cover
+        # Note token_urlsafe returns a 'length'-byte string which is then
+        # base64 encoded so is longer than length, so only return last
+        # 'length' characters.
+        return secrets.token_urlsafe(length)[:length]
 
 except ImportError:  # pragma: no cover
     import random
